@@ -59,6 +59,7 @@
 #include <dune/grid/common/gridenums.hh>
 
 #if HAVE_ECL_INPUT
+#include <opm/input/eclipse/EclipseState/Grid/EclipseGrid.hpp>
 #include <opm/input/eclipse/EclipseState/Grid/NNC.hpp>
 #endif
 
@@ -100,27 +101,6 @@ void refine_and_check(const Dune::cpgrid::Geometry<3, 3>&,
                       const std::array<int, 3>&,
                       bool);
 
-void refinePatch_and_check(const std::array<int,3>&,
-                           const std::array<int,3>&,
-                           const std::array<int,3>&);
-
-void refinePatch_and_check(Dune::CpGrid&,
-                           const std::vector<std::array<int,3>>&,
-                           const std::vector<std::array<int,3>>&,
-                           const std::vector<std::array<int,3>>&,
-                           const std::vector<std::string>&);
-
-void check_global_refine(const Dune::CpGrid&,
-                         const Dune::CpGrid&);
-
-void fieldProp_check(const Dune::CpGrid& grid, Opm::EclipseGrid eclGrid, const std::string& deck_string);
-
-void testInactiveCellsLgrs(const std::string&,
-                           const std::vector<std::array<int,3>>&,
-                           const std::vector<std::array<int,3>>&,
-                           const std::vector<std::array<int,3>>&,
-                           const std::vector<std::string>&);
-
 namespace Dune
 {
 namespace cpgrid
@@ -147,29 +127,6 @@ class CpGridData
     void ::refine_and_check(const Dune::cpgrid::Geometry<3, 3>&,
                             const std::array<int, 3>&,
                             bool);
-    friend
-    void ::refinePatch_and_check(const std::array<int,3>&,
-                                 const std::array<int,3>&,
-                                 const std::array<int,3>&);
-
-    friend
-    void ::refinePatch_and_check(Dune::CpGrid&,
-                                 const std::vector<std::array<int,3>>&,
-                                 const std::vector<std::array<int,3>>&,
-                                 const std::vector<std::array<int,3>>&,
-                                 const std::vector<std::string>&);
-
-    friend
-    void ::check_global_refine(const Dune::CpGrid&,
-                               const Dune::CpGrid&);
-    friend
-    void ::fieldProp_check(const Dune::CpGrid& grid, Opm::EclipseGrid eclGrid, const std::string& deck_string);
-    friend
-    void ::testInactiveCellsLgrs(const std::string&,
-                                 const std::vector<std::array<int,3>>&,
-                                 const std::vector<std::array<int,3>>&,
-                                 const std::vector<std::array<int,3>>&,
-                                 const std::vector<std::string>&);
 
 private:
     CpGridData(const CpGridData& g);
@@ -197,19 +154,20 @@ public:
 
     /// Constructor for parallel grid data.
     /// \param comm The MPI communicator
+    /// \param data Pointer to existing data to use
     /// Default constructor.
     explicit CpGridData(MPIHelper::MPICommunicator comm,  std::vector<std::shared_ptr<CpGridData>>& data);
 
 
-    
+
     /// Constructor
-    explicit CpGridData(std::vector<std::shared_ptr<CpGridData>>& data);  
+    explicit CpGridData(std::vector<std::shared_ptr<CpGridData>>& data);
     /// Destructor
     ~CpGridData();
-  
-   
-    
-    
+
+
+
+
     /// number of leaf entities per codim in this process
     int size(int codim) const;
 
@@ -224,61 +182,136 @@ public:
     }
 
     /// Read the Eclipse grid format ('grdecl').
-    /// \param filename the name of the file to read.
-    /// \param periodic_extension if true, the grid will be (possibly) refined, so that
-    ///        intersections/faces along i and j boundaries will match those on the other
-    ///        side. That is, i- faces will match i+ faces etc.
-    void readEclipseFormat(const std::string& filename, bool periodic_extension, bool turn_normals = false);
+    ///
+    /// \param[in] filename the name of the file to read.
+    ///
+    /// \param[in] periodic_extension Whether or not to process the
+    /// resulting grid in order to have intersections/faces along i and j
+    /// boundaries match those on the other side. That is, i- faces will
+    /// match i+ faces etc.
+    ///
+    /// \param[in] turn_normals Whether or not to turn all normals.  This is
+    /// intended for handling inputs with wrong orientations.
+    ///
+    /// \param[in] edge_conformal Whether or not to construct an
+    /// edge-conformal grid.  Typically useful in geo-mechanical
+    /// applications.
+    void readEclipseFormat(const std::string& filename,
+                           bool periodic_extension,
+                           bool turn_normals = false,
+                           bool edge_conformal = false);
 
 #if HAVE_ECL_INPUT
     /// Read the Eclipse grid format ('grdecl').
-    /// \param deck the parsed deck from opm-parser (which is a low-level object)
-    /// \param periodic_extension if true, the grid will be (possibly) refined, so that
-    ///        intersections/faces along i and j boundaries will match those on the other
-    ///        side. That is, i- faces will match i+ faces etc.
-    /// \param turn_normals if true, all normals will be turned. This is intended for handling inputs with wrong orientations.
-    /// \param clip_z if true, the grid will be clipped so that the top and bottom will be planar.
-    /// \param poreVolume pore volumes for use in MINPV processing, if asked for in deck
-    void processEclipseFormat(const Opm::Deck& deck, bool periodic_extension, bool turn_normals = false, bool clip_z = false,
-                              const std::vector<double>& poreVolume = std::vector<double>());
+    ///
+    /// \param[in] deck Low-level input Deck object from the OPM Parser.
+    ///
+    /// \param[in] periodic_extension Whether or not to process the
+    /// resulting grid in order to have intersections/faces along i and j
+    /// boundaries match those on the other side. That is, i- faces will
+    /// match i+ faces etc.
+    ///
+    /// \param[in] turn_normals Whether or not to turn all normals.  This is
+    /// intended for handling inputs with wrong orientations.
+    ///
+    /// \param[in] clip_z Whether or not to clip the result grid in
+    /// order to have planar top and bottom surfaces.
+    ///
+    /// \param[in] poreVolume pore volumes for use in MINPV processing, if
+    /// asked for in deck
+    ///
+    /// \param[in] edge_conformal Whether or not to construct an
+    /// edge-conformal grid.  Typically useful in geo-mechanical
+    /// applications.
+    void processEclipseFormat(const Opm::Deck& deck,
+                              bool periodic_extension,
+                              bool turn_normals = false,
+                              bool clip_z = false,
+                              const std::vector<double>& poreVolume = std::vector<double>{},
+                              bool edge_conformal = false);
 
     /// Read the Eclipse grid format ('grdecl').
-    /// \param ecl_grid the high-level object from opm-parser which represents the simulation's grid
-    ///        In a parallel run this may be a nullptr on all ranks but rank zero.
-    /// \param ecl_state the object from opm-parser provide information regarding to pore volume, NNC,
-    ///        aquifer information when ecl_state is available. NNC and aquifer connection
-    ///        information will also be updated during the function call when available and necessary.
-    /// \param periodic_extension if true, the grid will be (possibly) refined, so that
-    ///        intersections/faces along i and j boundaries will match those on the other
-    ///        side. That is, i- faces will match i+ faces etc.
-    /// \param turn_normals if true, all normals will be turned. This is intended for handling inputs with wrong orientations.
-    /// \param clip_z if true, the grid will be clipped so that the top and bottom will be planar.
-    /// \param pichActive Whether PINCH keyword was specified
-    /// \return A vector of removed cells and created NNCs with transmissibilities die to PINCH item 4 all.
+    ///
+    /// \param[in] ecl_grid Simulation's grid.  In a parallel run this may
+    /// be a nullptr on all ranks other than rank zero.
+    ///
+    /// \param[in,out] ecl_state High level object from opm-common that
+    /// provides information regarding pore volumes, NNCs, and aquifers.
+    /// NNC and aquifer connection information will also be updated during
+    /// the function call when necessary if \p ecl_state is non-null.
+    ///
+    /// \param[in] periodic_extension Whether or not to process the
+    /// resulting grid in order to have intersections/faces along i and
+    /// j boundaries match those on the other side. That is, i- faces
+    /// will match i+ faces etc.
+    ///
+    /// \param[in] turn_normals Whether or not to turn all normals.
+    /// This is intended for handling inputs with wrong orientations.
+    ///
+    /// \param[in] clip_z Whether or not to clip the result grid in
+    /// order to have planar top and bottom surfaces.
+    ///
+    /// \param[in] pinchActive Whether or not to force specific pinch
+    /// behaviour.  If set, a face will connect two vertical cells, that
+    /// are topological connected, even if there are cells with zero
+    /// volume between them. If false these cells will not be connected
+    /// despite their faces coinciding.
+    ///
+    /// \param[in] edge_conformal Whether or not to construct an
+    /// edge-conformal grid.  Typically useful in geo-mechanical
+    /// applications.
+    ///
+    /// \return Cells removed due low pore-volume and across which to create
+    /// non-neighbouring connections if item 4 of the 'PINCH' keyword is set
+    /// to 'ALL'.
     std::vector<std::size_t>
-    processEclipseFormat(const Opm::EclipseGrid* ecl_grid, Opm::EclipseState* ecl_state,
-                         bool periodic_extension, bool turn_normals = false, bool clip_z = false,
-                         bool pinchActive = true);
+    processEclipseFormat(const Opm::EclipseGrid* ecl_grid,
+                         Opm::EclipseState* ecl_state,
+                         bool periodic_extension,
+                         bool turn_normals = false,
+                         bool clip_z = false,
+                         bool pinchActive = true,
+                         bool edge_conformal = false);
 #endif
 
     /// Read the Eclipse grid format ('grdecl').
-    /// \param input_data the data in grdecl format, declared in preprocess.h.
     ///
-    /// \param ecl_state the object from opm-parser provide information regarding to pore volume, NNC,
-    ///        aquifer information when ecl_state is available. NNC and aquifer connection
-    ///        information will also be updated during the function call when available and necessary.
-    /// \param nnc is the non-neighboring connections
-    /// \param remove_ij_boundary if true, will remove (i, j) boundaries. Used internally.
-    /// \param pinchActive If true, we will add faces between vertical cells that have only inactive cells or cells
-    ///            with zero volume between them. If false these cells will not be connected.
-    /// \param tolerance_unique_points Tolerance used to identify points based on their cooridinate
+    /// \param[in] input_data Corner-point grid input data.
+    ///
+    /// \param[in,out] ecl_state High level object from opm-common that
+    /// provides information regarding pore volumes, NNCs and aquifers.  NNC
+    /// and aquifer connection information will also be updated during the
+    /// function call when necessary if \p ecl_state is non-null.
+    ///
+    /// \param[in,out] nnc Non-neighboring connections.
+    ///
+    /// \param[in] remove_ij_boundary True to remove outer cell layer
+    ///
+    /// \param[in] turn_normals Whether or not to turn all normals.
+    /// This is intended for handling inputs with wrong orientations.
+    ///
+    /// \param[in] pinchActive Whether or not to force specific pinch
+    /// behaviour.  If set, a face will connect two vertical cells, that are
+    /// topological connected, even if there are cells with zero volume
+    /// between them. If false these cells will not be connected despite
+    /// their faces coinciding.
+    ///
+    /// \param[in] tolerance_unique_points Tolerance used to identify points
+    /// based on their cooridinate.
+    ///
+    /// \param[in] edge_conformal Whether or not to construct an
+    /// edge-conformal grid.  Typically useful in geo-mechanical
+    /// applications.
     void processEclipseFormat(const grdecl& input_data,
 #if HAVE_ECL_INPUT
                               Opm::EclipseState* ecl_state,
 #endif
                               std::array<std::set<std::pair<int, int>>, 2>& nnc,
-                              bool remove_ij_boundary, bool turn_normals, bool pinchActive,
-                              double tolerance_unique_points);
+                              bool remove_ij_boundary,
+                              bool turn_normals,
+                              bool pinchActive,
+                              double tolerance_unique_points,
+                              bool edge_conformal);
 
     /// @brief
     ///    Extract Cartesian index triplet (i,j,k) of an active cell.
@@ -287,19 +320,68 @@ public:
     ///    Active cell index.
     ///
     /// @param [out] ijk  Cartesian index triplet
-    void getIJK(int c, std::array<int,3>& ijk) const
-    {
-        ijk = getIJK(global_cell_[c], logical_cartesian_size_);
-    }
+    void getIJK(int c, std::array<int,3>& ijk) const;
 
     int cellFace(int cell, int local_index) const
     {
         return cell_to_face_[cpgrid::EntityRep<0>(cell, true)][local_index].index();
     }
 
+    auto cellToFace(int cellIdx) const
+    {
+        return cell_to_face_[cpgrid::EntityRep<0>(cellIdx, true)];
+    }
+
+    const auto& cellToPoint() const
+    {
+        return cell_to_point_;
+    }
+
+    const auto& cellToPoint(int cellIdx) const
+    {
+        return cell_to_point_[cellIdx];
+    }
+
     int faceToCellSize(int face) const {
-        Dune::cpgrid::EntityRep<1> faceEntity(face, true);
-        return face_to_cell_[faceEntity].size();
+        Dune::cpgrid::EntityRep<1> faceRep(face, true);
+        return face_to_cell_[faceRep].size();
+    }
+
+    auto faceTag(int faceIdx) const
+    {
+        Dune::cpgrid::EntityRep<1> faceRep(faceIdx, true);
+        return face_tag_[faceRep];
+    }
+
+    auto faceNormals(int faceIdx) const
+    {
+        Dune::cpgrid::EntityRep<1> faceRep(faceIdx, true);
+        return face_normals_[faceRep];
+    }
+
+    auto faceToPoint(int faceIdx) const
+    {
+        return face_to_point_[faceIdx];
+    }
+
+    int numFaces() const
+    {
+        return face_to_cell_.size();
+    }
+
+    auto cornerHistorySize() const
+    {
+        return corner_history_.size();
+    }
+
+    const auto& getCornerHistory(int cornerIdx) const
+    {
+        if(cornerHistorySize()) {
+            return corner_history_[cornerIdx];
+        }
+        else {
+            OPM_THROW(std::logic_error, "Vertex has no history record.\n");
+        }
     }
 
     /// Return global_cell_ of any level grid, or the leaf grid view (in presence of refinement).
@@ -313,69 +395,10 @@ public:
         return  global_cell_;
     }
 
-    /// @brief Extract Cartesian index triplet (i,j,k) given an index between 0 and NXxNYxNZ -1
-    ///    where NX, NY, and NZ is the total amoung of cells in each direction x-,y-,and z- respectively.
-    ///
-    /// @param [in] idx      Integer between 0 and cells_per_dim[0]*cells_per_dim[1]*cells_per_dim[2]-1
-    /// @param [in] cells_per_dim
-    /// @return Cartesian index triplet.
-    std::array<int,3> getIJK(int idx_in_parent_cell, const std::array<int,3>& cells_per_dim) const
-    {
-        // idx = k*cells_per_dim_[0]*cells_per_dim_[1] + j*cells_per_dim_[0] + i
-        // with 0<= i < cells_per_dim_[0], 0<= j < cells_per_dim_[1], 0<= k <cells_per_dim_[2].
-        assert(cells_per_dim[0]);
-        assert(cells_per_dim[1]);
-        assert(cells_per_dim[2]);
-
-        std::array<int,3> ijk = {0,0,0};
-        ijk[0] = idx_in_parent_cell % cells_per_dim[0]; idx_in_parent_cell /= cells_per_dim[0];
-        ijk[1] = idx_in_parent_cell % cells_per_dim[1];
-        ijk[2] = idx_in_parent_cell /cells_per_dim[1];
-        return ijk;
-    }
-
-    /// @brief Determine if a finite amount of patches (of cells) are disjoint, namely, they do not share any corner nor face.
-    ///
-    /// @param [in]  startIJK_vec  Vector of Cartesian triplet indices where each patch starts.
-    /// @param [in]  endIJK_vec    Vector of Cartesian triplet indices where each patch ends.
-    ///                            Last cell part of the lgr will be {endIJK_vec[<patch>][0]-1, ... ,endIJK_vec[<patch>][2]-1}.
-    bool disjointPatches(const std::vector<std::array<int,3>>& startIJK_vec, const std::vector<std::array<int,3>>& endIJK_vec) const;
-
-    /// @brief Compute cell indices of selected patches of cells (Cartesian grid required).
-    ///
-    /// @param [in]  startIJK_vec  Vector of Cartesian triplet indices where each patch starts.
-    /// @param [in]  endIJK_vec    Vector of Cartesian triplet indices where each patch ends.
-    ///                            Last cell part of the lgr will be {endIJK_vec[<patch>][0]-1, ... endIJK_vec[<patch>][2]-1}.
-    ///
-    /// @return allPatches_cells
-    std::vector<int>
-    getPatchesCells(const std::vector<std::array<int,3>>& startIJK_vec, const std::vector<std::array<int,3>>& endIJK_vec) const;
-
     /// @brief Check all cells selected for refinement have no NNCs (no neighbor connections).
     ///        Assumption: all grid cells are active.
     bool hasNNCs(const std::vector<int>& cellIndices) const;
 
-    /// @brief Check startIJK and endIJK of each patch of cells to be refined are valid, i.e.
-    ///        startIJK and endIJK vectors have the same size and, startIJK < endIJK coordenate by coordenate.
-    ///
-    /// @param [in]  startIJK_vec       Vector of Cartesian triplet indices where each patch starts.
-    /// @param [in]  endIJK_vec         Vector of Cartesian triplet indices where each patch ends.
-    ///                                 Last cell part of the lgr will be {endIJK_vec[patch][0]-1, ..., endIJK_vec[patch][2]-1}.
-    void validStartEndIJKs(const std::vector<std::array<int,3>>& startIJK_vec, const std::vector<std::array<int,3>>& endIJK_vec) const;
-
-    /// @brief Check that every cell to be refined has cuboid shape.
-    void checkCuboidShape(const std::vector<int>& cellIdx_vec) const;
-
-    /// @brief Determine if a finite amount of patches (of cells) share a face.
-    ///
-    /// @param [in]  startIJK_vec  Vector of Cartesian triplet indices where each patch starts.
-    /// @param [in]  endIJK_vec    Vector of Cartesian triplet indices where each patch ends.
-    ///                            Last cell part of the lgr will be {endIJK_vec[<patch>][0]-1, ... ,endIJK_vec[<patch>][2]-1}.
-    bool patchesShareFace(const std::vector<std::array<int,3>>& startIJK_vec, const std::vector<std::array<int,3>>& endIJK_vec) const;
-
-    int sharedFaceTag(const std::vector<std::array<int,3>>& startIJK_2Patches, const std::vector<std::array<int,3>>& endIJK_2Patches) const;
-
-    
     /// @brief Mark entity for refinement or coarsening.
     ///
     /// Refinement on CpGrid is partially supported for Cartesian grids, with the keyword CARFIN.
@@ -397,6 +420,13 @@ public:
 
     /// @brief Set mightVanish flags for elements that will be refined in the next adapt() call
     ///        Need to be called after elements have been marked for refinement.
+    ///
+    ///        Communicate marks accross processes, in parallel runs.
+    ///        An element may be marked somewhere in opm-simulators because it does not fulfill a
+    ///        certain property, regardless of whether it belongs to the interior or overlap
+    ///        partition.
+    ///
+    /// @return True if at least one element has been marked for refinement, false otherwise.
     bool preAdapt();
 
     /// TO DO: Documentation. Triggers the grid refinement process - Currently, returns preAdapt()
@@ -406,109 +436,8 @@ public:
     void postAdapt();
 
 private:
-    /// @brief Check compatibility of number of subdivisions of neighboring LGRs.
-    ///
-    /// Check shared faces on boundaries of LGRs. Not optimal since the code below does not take into account
-    /// active/inactive cells, instead, relies on "ijk-computations".
-    ///
-    /// @param [in]  cells_per_dim_vec    Vector of expected subdivisions per cell, per direction, in each LGR.
-    /// @param [in]  startIJK_vec         Vector of Cartesian triplet indices where each patch starts.
-    /// @param [in]  endIJK_vec           Vector of Cartesian triplet indices where each patch ends.
-    ///                                   Last cell part of the lgr will be {endIJK_vec[patch][0]-1, ..., endIJK_vec[patch][2]-1}.
-    /// @return True if all block of cells either do not share faces on their boundaries, or they may share faces with compatible
-    ///         subdivisions. Example: block1 and block2 share an I_FACE, then number of subdivisions NY NZ should coincide, i.e.
-    ///         if block1, block2 cells_per_dim values are {NX1, NY1, NZ1}, {NX2, NY2, NZ2}, respectively, then NY1 == NY2 and
-    ///         NZ1 == Nz2.
-    ///         False if at least two blocks share a face and their subdivions are not compatible. In the example above,
-    ///         if NY1 != NY2 or NZ1 != NZ2.
-    bool compatibleSubdivisions(const std::vector<std::array<int,3>>& cells_per_dim_vec,
-                                const std::vector<std::array<int,3>>& startIJK_vec,
-                                const std::vector<std::array<int,3>>& endIJK_vec) const;
-
     std::array<Dune::FieldVector<double,3>,8> getReferenceRefinedCorners(int idx_in_parent_cell, const std::array<int,3>& cells_per_dim) const;
 
-    /// @brief Compute amount of cells in each direction of a patch of cells. (Cartesian grid required).
-    ///
-    /// @param [in]  startIJK  Cartesian triplet index where the patch starts.
-    /// @param [in]  endIJK    Cartesian triplet index where the patch ends.
-    ///                        Last cell part of the lgr will be {endijk[0]-1, ... endIJK[2]-1}.
-    ///
-    /// @return patch_dim Patch dimension {#cells in x-direction, #cells in y-direction, #cells in z-direction}.
-    std::array<int,3> getPatchDim(const std::array<int,3>& startIJK, const std::array<int,3>& endIJK) const;
-
-    /// @brief Compute corner indices of a patch of cells (Cartesian grid required).
-    ///
-    /// @param [in]  startIJK  Cartesian triplet index where the patch starts.
-    /// @param [in]  endIJK    Cartesian triplet index where the patch ends.
-    ///                        Last cell part of the lgr will be {endijk[0]-1, ... endIJK[2]-1}.
-    ///
-    /// @return patch_corners
-    std::vector<int> getPatchCorners(const std::array<int,3>& startIJK, const std::array<int,3>& endIJK) const;
-
-    /// @brief Compute face indices of a patch of cells (Cartesian grid required).
-    ///
-    /// @param [in]  startIJK  Cartesian triplet index where the patch starts.
-    /// @param [in]  endIJK    Cartesian triplet index where the patch ends.
-    ///                        Last cell part of the lgr will be {endijk[0]-1, ... endIJK[2]-1}.
-    ///
-    /// @return patch_faces
-    std::vector<int> getPatchFaces(const std::array<int,3>& startIJK, const std::array<int,3>& endIJK) const;
-
-    /// @brief Compute cell indices of a patch of cells (Cartesian grid required).
-    ///
-    /// @param [in]  startIJK  Cartesian triplet index where the patch starts.
-    /// @param [in]  endIJK    Cartesian triplet index where the patch ends.
-    ///                        Last cell part of the lgr will be {endIJK[0]-1, ... endIJK[2]-1}.
-    ///
-    /// @return patch_cells
-    std::vector<int> getPatchCells(const std::array<int,3>& startIJK, const std::array<int,3>& endIJK) const;
-
-    /// @brief Compute patch boundary corner indices (Cartesian grid required).
-    ///
-    /// @param [in]  startIJK  Cartesian triplet index where the patch starts.
-    /// @param [in]  endIJK    Cartesian triplet index where the patch ends.
-    ///                        Last cell part of the lgr will be {endijk[0]-1, ... endIJK[2]-1}.
-    ///
-    /// @return patch_boundary_corners
-    std::vector<int> getPatchBoundaryCorners(const std::array<int,3>& startIJK, const std::array<int,3>& endIJK) const;
-
-    /// @brief Compute patch boundary face indices (Cartesian grid required).
-    ///
-    /// @param [in]  startIJK  Cartesian triplet index where the patch starts.
-    /// @param [in]  endIJK    Cartesian triplet index where the patch ends.
-    ///                        Last cell part of the lgr will be {endijk[0]-1, ... endIJK[2]-1}.
-    ///
-    /// @return patch_boundary_faces
-    std::array<std::vector<int>,6> getBoundaryPatchFaces(const std::array<int,3>& startIJK, const std::array<int,3>& endIJK) const;
-
-    /// @brief For selected cell indices, computes the variation in x-,y-, and z-direction, assuming each cell has cubiod shape.
-    std::array<std::vector<double>,3> getWidthsLengthsHeights(const std::array<int,3>& startIJK, const std::array<int,3>& endIJK) const;
-
-    /// @brief Construct a 'fake cell (Geometry<3,3> object)' out of a patch of cells.(Cartesian grid required).
-    ///
-    /// cellifyPatch() builds a Geometry<3,3> object, 'a celliFIED patch', from a connected patch formed
-    /// by the product of consecutive cells in each direction; selecting 8 corners of the patch boundary,
-    /// computing center and volume.
-    ///
-    /// @param [in] startIJK                   Cartesian triplet index where the patch starts.
-    /// @param [in] endIJK                     Cartesian triplet index where the patch ends.
-    /// @param [in] patch_cells                Cell indices from the block-shaped patch.
-    /// @param [out] cellifiedPatch_geometry   Required as an argument when creating a Geomtry<3,3> object.
-    /// @param [out] cellifiedPatch_to_point   To store the 8 corners of the created cellifiedPatch.
-    /// @param [out] allcorners_cellifiedPatch Required to build a Geometry<3,3> object.
-    ///
-    /// @return 'cellifiedPatchCell'         Geometry<3,3> object.
-    Geometry<3,3> cellifyPatch(const std::array<int,3>& startIJK, const std::array<int,3>& endIJK,
-                               const std::vector<int>& patch_cells, DefaultGeometryPolicy& cellifiedPatch_geometry,
-                               std::array<int,8>& cellifiedPatch_to_point,
-                               std::array<int,8>& allcorners_cellifiedPatch) const;
-
-    // @brief Compute the average of array<double,3>.
-    //
-    // @param [in] vector of array<double,3>
-    // @return     array<double,3> (average of the entries of the given vector).
-    std::array<double,3> getAverageArr(const std::vector<std::array<double,3>>& vec) const;
-
 public:
     /// Add doc/or remove method and replace it with better approach
     int getGridIdx() const {
@@ -547,6 +476,20 @@ public:
     const std::vector<std::tuple<int,std::vector<int>>>& getParentToChildren() const {
         return parent_to_children_cells_;
     }
+
+    const cpgrid::DefaultGeometryPolicy getGeometry() const
+    {
+        return geometry_;
+    }
+
+    int getLeafIdxFromLevelIdx(int level_cell_idx) const
+    {
+        if (level_to_leaf_cells_.empty()) {
+            OPM_THROW(std::logic_error, "Grid has no LGRs. No mapping to the leaf.\n");
+        }
+        return level_to_leaf_cells_[level_cell_idx];
+    }
+
     /// @brief Refine a single cell and return a shared pointer of CpGridData type.
     ///
     /// refineSingleCell() takes a cell and refines it in a chosen amount of cells (per direction); creating the
@@ -576,27 +519,6 @@ public:
                 const std::vector<std::array<int,2>>>                // child_to_parent_cells
     refineSingleCell(const std::array<int,3>& cells_per_dim, const int& parent_idx) const;
 
-    /// @brief Refine a (connected block-shaped) patch of cells. Based on the patch, a Geometry<3,3> object is created and refined.
-    ///
-    /// @param [in] cells_per_dim            Number of (refined) cells in each direction that each parent cell should be refined to.
-    /// @param [in] startIJK                 Cartesian triplet index where the patch starts.
-    /// @param [in] endIJK                   Cartesian triplet index where the patch ends.
-    ///                                      Last cell part of the lgr will be {endijk[0]-1, ... endIJK[2]-1}.
-    ///
-    /// @return refined_grid_ptr                   Shared pointer of CpGridData type, pointing at the refined_grid
-    /// @return boundary_old_to_new_corners/faces  Corner/face indices on the patch-boundary associated with new-born-entity indices.
-    /// @return parent_to_children_faces/cell      For each parent face/cell, we store its child-face/cell indices.
-    ///                                            {parent face/cell index in coarse level, {indices of its children in refined level}}
-    /// @return child_to_parent_faces/cells        {child index, parent index}
-    std::tuple< std::shared_ptr<CpGridData>,
-                const std::vector<std::array<int,2>>,                // boundary_old_to_new_corners
-                const std::vector<std::tuple<int,std::vector<int>>>, // boundary_old_to_new_faces
-                const std::vector<std::tuple<int,std::vector<int>>>, // parent_to_children_faces
-                const std::vector<std::tuple<int,std::vector<int>>>, // parent_to_children_cell
-                const std::vector<std::array<int,2>>,                // child_to_parent_faces
-                const std::vector<std::array<int,2>>>                // child_to_parent_cells
-    refinePatch(const std::array<int,3>& cells_per_dim, const std::array<int,3>& startIJK, const std::array<int,3>& endIJK) const;
-
     // @breif Compute center of an entity/element/cell in the Eclipse way:
     //        - Average of the 4 corners of the bottom face.
     //        - Average of the 4 corners of the top face.
@@ -895,10 +817,10 @@ private:
     /** Copy of (CpGrid object).data_ associated with the CpGridData object. */
     std::vector<std::shared_ptr<CpGridData>>* level_data_ptr_;
     // SUITABLE FOR ALL LEVELS EXCEPT FOR LEAFVIEW
-    /** Map between level and leafview cell indices. Only cells (from that level) that appear in leafview count. -1 when the cell vanished.*/  
+    /** Map between level and leafview cell indices. Only cells (from that level) that appear in leafview count. -1 when the cell vanished.*/
     std::vector<int> level_to_leaf_cells_; // In entry 'level cell index', we store 'leafview cell index'
     /** Parent cells and their children. Entry is {-1, {}} when cell has no children.*/ // {level LGR, {child0, child1, ...}}
-    std::vector<std::tuple<int,std::vector<int>>> parent_to_children_cells_; 
+    std::vector<std::tuple<int,std::vector<int>>> parent_to_children_cells_;
     /** Amount of children cells per parent cell in each direction. */ // {# children in x-direction, ... y-, ... z-}
     std::array<int,3> cells_per_dim_;
     // SUITABLE ONLY FOR LEAFVIEW
@@ -912,8 +834,8 @@ private:
     /** Level-grid or Leaf-grid cell to parent cell and refined-cell-in-parent-cell index (number between zero and total amount
         of children per parent (cells_per_dim[0]_*cells_per_dim_[1]*cells_per_dim_[2])). Entry is -1 when cell has no father. */
     std::vector<int> cell_to_idxInParentCell_;
-    
-
+    /** To keep track of refinement processes */
+    int refinement_max_level_{0};
 
     /// \brief Object for collective communication operations.
     Communication ccobj_;
