@@ -39,6 +39,7 @@
 
 #include <dune/common/version.hh>
 #include <dune/geometry/type.hh>
+#include <dune/geometry/referenceelements.hh>
 #include <dune/grid/common/gridenums.hh>
 
 #include "PartitionTypeIndicator.hpp"
@@ -82,21 +83,19 @@ class Entity : public EntityRep<codim>
 public:
     /// @brief
     /// @todo Doc me!
-    enum { codimension = codim };
-    enum { dimension = 3 };
-    enum { mydimension = dimension - codimension };
-    enum { dimensionworld = 3 };
+    static constexpr int codimension = codim;
+    static constexpr int dimension = 3;
+    static constexpr int mydimension = dimension - codimension;
+    static constexpr int dimensionworld = 3;
 
     // the official DUNE names
     typedef Entity    EntitySeed;
 
-    /// @brief
-    /// @todo Doc me!
-    /// @tparam
+    /** \brief Export supported entity types */
     template <int cd>
     struct Codim
     {
-        typedef cpgrid::Entity<cd> Entity;
+        using Entity = ::Dune::cpgrid::Entity<cd>;
     };
 
 
@@ -183,13 +182,10 @@ public:
 
     /// @brief In serial run, the only partitionType() is InteriorEntity.
     ///        Only needed when distributed_data_ is not empty.
-    PartitionType partitionType() const;
-
-    /// @brief For parallel run, the entity - for now - does not see the CpGrid therefore we pass a bool to make
-    ///        the entity aware of the fact the the grid has been distributed.
-    ///        Each cell inherits the partition type of its origin (either parent cell or equivalent cell in level 0).
+    ///        In parallel run, the entity each cell inherits the partition type of its origin
+    ///        (either parent cell or equivalent cell in level 0).
     ///        Only needed when distributed_data_ is not empty.
-    PartitionType partitionTypeWhenLgrs(bool) const;
+    PartitionType partitionType() const;
 
     /// @brief Return marker object (GeometryType object) representing the reference element of the entity.
     ///        Currently, cube type for all entities (cells and vertices).
@@ -225,15 +221,19 @@ public:
     /// @brief Iterator end over the children/beyond last child iterator.
     HierarchicIterator hend(int) const;
 
-    /// \brief Returns true, if the entity has been created during the last call to adapt(). Dummy.
+    /// \brief Returns true, if the entity has been created during the last call to adapt().
+    ///
+    /// Note: An element created during the last refinement step may still be refined further
+    /// on a higher level (e.g., through nested refinement).
+    /// The isNew flag is used to identify such newly created elements so that data
+    /// interpolation is applied only to them.
     bool isNew() const;
 
-    /// \brief Returns true, if entity might disappear during the next call to adapt().
+    /// \brief Indicates whether the entity may be removed in the next call to adapt().
     ///
-    ///        Currently, returns true if the element has been marked to be refined. Therefore, it "vanishes"
-    ///        in the sense that it gets replaced by its children.
-    ///        To be modified: When supporting coarsening, mightVanish should return true for all the "siblings"
-    ///        refined cells of one refined cell that got marked to be coarsened.
+    /// For CpGrid, this currently always returns false, since only refinement
+    /// (and no coarsening) is supported. A return value of false guarantees
+    /// that the entity will still exist after adaptation.
     bool mightVanish() const;
 
     /// @brief ONLY FOR CELLS (Entity<0>)
@@ -277,35 +277,35 @@ public:
     /// \return return true if seed is pointing to a valid entity
     bool isValid () const;
 
-    /// \brief Returns (1) parent entity in the level-grid the parent cell was born, if the entity was born in any LGR.
+    /// \brief Returns (1) oldest ancestor, i.e., oldest parent entity in the level-grid 0, if the entity was born in any refined level
+    ///                    grid with level>0.
     ///                (2) if the entity has no father, and is a leaf-grid-view entity, it returns the equivalent element on the level
     ///                    that it was born. Namely, for coarse cells on the leaf never involved in any refinement process, we get the
-    ///                    equivalent entity in the GLOBAL grid (level 0). Notice that if it's a refined cell on the leaf, it does have a
-    ///                    father, in that case, this method returns the father entity.
-    ///                (3) Otherwise, returns itself. Notice that if the element is a refined one, this method returns the father() entity.
-    ///                    For a coarse cell never involved in any refinement, if it's also a leaf-grid-cell, it returns the equivalent
-    ///                    cell from level 0. Otherwise, it returns itself, which means that its grid is level 0 and the entity itself is already
+    ///                    equivalent entity in the GLOBAL grid (level 0).
+    ///                (3) Otherwise, returns itself, which means that its grid is the level 0 and the entity itself is already
     ///                    its "origin".
     Entity<0> getOrigin() const;
 
-    /// \brief To be invoked only for leaf-grid-view entities. Get equivalent element on the level grid the leaf-entity was born.
-    Entity<0> getLevelElem() const;
-
     /// \brief Get equivalent element on the level grid where the entity was born, if grid = leaf-grid-view. Otherwise, return itself.
-    Entity<0> getEquivLevelElem() const;
+    Entity<0> getLevelElem() const;
 
     /// \brief Get Cartesian Index in the level grid view where the Entity was born.
     int getLevelCartesianIdx() const;
 
+    int getIdxInParentCell() const;
+
 protected:
     const CpGridData* pgrid_;
-private:
-    // static    to not need any extra storage per Enitity. One object used for all instances
-    // constexpr to allow for in-class instantiation
-    /// \brief Indices of corners in entity's geometry in father reference element, for the Geometry returned by geometryInFather
-    static constexpr std::array<int,8> in_father_reference_elem_corner_indices_ = {0,1,2,3,4,5,6,7};
 };
 
+// Reference element for a given entity
+template <int codim>
+auto referenceElement(const Entity<codim>&)
+{
+    return Dune::referenceElement<double,3-codim>(Dune::GeometryTypes::cube(3-codim));
+}
+
+
 } // namespace cpgrid
 } // namespace Dune
 
@@ -367,11 +367,6 @@ PartitionType Entity<codim>::partitionType() const
     return pgrid_->partition_type_indicator_->getPartitionType(*this);
 }
 
-template <int codim>
-PartitionType Entity<codim>::partitionTypeWhenLgrs(bool lgrsOnDistributedGrid) const
-{
-    return pgrid_->partition_type_indicator_->getPartitionTypeWhenLgrs(*this, lgrsOnDistributedGrid);
-}
 } // namespace cpgrid
 } // namespace Dune
 
@@ -470,33 +465,35 @@ int Entity<codim>::level() const
 template<int codim>
 bool Entity<codim>::isLeaf() const
 {
-    if (pgrid_ -> parent_to_children_cells_.empty()){ // LGR cells
+    if (pgrid_ -> parent_to_children_cells_.empty()){ // Grid cells without nested refinement
         return true;
     }
     else {
-        return (std::get<0>((pgrid_ -> parent_to_children_cells_)[this-> index()]) == -1);  // Cells from GLOBAL, not involved in any LGR
+        return (std::get<0>((pgrid_ -> parent_to_children_cells_)[this-> index()]) == -1);  // Not involved in any LGR
     }
 }
 
 template<int codim>
 bool Entity<codim>::isNew() const
 {
-    // WIP
-    // For an Entity "to be new" means that
-    // - it has a father
-    // - it has been created in the last call of adapt()
-    // To determine that, we track the level of the Entity and
-    // check if it was marked for refinement in the previos state
-    // of current_view_data_
-    return (isLeaf() && hasFather());
+    int data_count = pgrid_->level_data_ptr_->size();
+    if (data_count == 1) { // CpGrid has only level zero grid.
+        return false; // no element is new
+    }
+    else {
+        assert(data_count >= 2); // Level zero and leaf grids, plus at least one refined level grid
+        // In nested refinement, an element created during the last refinement step
+        // may itself be refined further within the same step at a higher level.
+        // To handle this case, we also check whether the element is a leaf
+        // (i.e., it has no children).
+        return isLeaf() && (pgrid_->refinement_max_level_ == data_count - 2); // minus "level zero" and "leaf"
+    }
 }
 
-
 template<int codim>
 bool Entity<codim>::mightVanish() const
 {
-    const auto refinementMark = pgrid_ -> getMark(*this);
-    return (refinementMark == 1);
+    return false;
 }
 
 template<int codim>
@@ -523,12 +520,25 @@ Entity<0> Entity<codim>::father() const
     }
 }
 
+template<int codim>
+int Dune::cpgrid::Entity<codim>::getIdxInParentCell() const
+{
+    return pgrid_ -> cell_to_idxInParentCell_[this->index()];
+}
+
+
 template<int codim>
 Dune::cpgrid::Geometry<3,3> Dune::cpgrid::Entity<codim>::geometryInFather() const
 {
     if (!(this->hasFather())){
         OPM_THROW(std::logic_error, "Entity has no father.");
     }
+
+    // Indices of corners in entity's geometry in father reference element.
+    static constexpr std::array<int,8> in_father_reference_elem_corner_indices = {0,1,2,3,4,5,6,7};
+    // 'static': The returned object Geometry<3,3> stores a pointer to in_father_reference_elem_corner_indices. Therefore,
+    // this variable is declared static to prolongate its lifetime beyond this function (static storage duration).
+
     auto idx_in_parent_cell = pgrid_ -> cell_to_idxInParentCell_[this->index()];
     if (idx_in_parent_cell !=-1) {
         const auto& cells_per_dim =  (*(pgrid_ -> level_data_ptr_))[this->level()] -> cells_per_dim_;
@@ -552,7 +562,7 @@ Dune::cpgrid::Geometry<3,3> Dune::cpgrid::Entity<codim>::geometryInFather() cons
         double volume_in_father_reference_elem = double(1)/(cells_per_dim[0]*cells_per_dim[1]*cells_per_dim[2]);
         // Construct (and return) the Geometry<3,3> of 'child-cell in the reference element of its father (unit cube)'.
         return Dune::cpgrid::Geometry<3,3>(center_in_father_reference_elem, volume_in_father_reference_elem,
-                                           in_father_reference_elem_corners, in_father_reference_elem_corner_indices_.data());
+                                           in_father_reference_elem_corners, in_father_reference_elem_corner_indices.data());
     }
     else {
         OPM_THROW(std::logic_error, "Entity has no father.");
@@ -562,18 +572,24 @@ Dune::cpgrid::Geometry<3,3> Dune::cpgrid::Entity<codim>::geometryInFather() cons
 template<int codim>
 Dune::cpgrid::Entity<0> Dune::cpgrid::Entity<codim>::getOrigin() const
 {
-    if (hasFather())
-    {
-        return this->father();
+    if (hasFather()) { // Entit is a refined cell belonging to
+        // a refined level grid (level>0) or to the leaf grid.
+        auto ancestor = this->father();
+        while (ancestor.hasFather()){
+            ancestor = ancestor.father();
+        }
+        assert(ancestor.level() == 0);
+        return ancestor;
     }
-    if (!(pgrid_ -> leaf_to_level_cells_.empty())) // entity on the LeafGridView
-    {  // leaf_to_level_cells_ [leaf idx] = { level where entity was born, cell idx in that level}
-        const int& levelElem = pgrid_->leaf_to_level_cells_[this->index()][0];
+    else if (!(pgrid_ -> leaf_to_level_cells_.empty())) { // Entity is a coarse cell
+        // (born in level zero grid, never involved in refinement) belonging to the leaf grid.
+        // leaf_to_level_cells_ [leaf idx] = { level where entity was born, cell idx in that level}
+        const int& level = pgrid_->leaf_to_level_cells_[this->index()][0];
+        assert(level == 0);
         const int& levelElemIdx = pgrid_->leaf_to_level_cells_[this->index()][1];
-        return Dune::cpgrid::Entity<0>( *((*(pgrid_ -> level_data_ptr_))[levelElem].get()), levelElemIdx, true);
+        return Dune::cpgrid::Entity<0>( *((*(pgrid_ -> level_data_ptr_))[level].get()), levelElemIdx, true);
     }
-    else
-    {
+    else { // Entity belongs to level zero grid.
         return *this;
     }
 }
@@ -589,22 +605,6 @@ Dune::cpgrid::Entity<0> Dune::cpgrid::Entity<codim>::getLevelElem() const
         const int& entityLevelIdx = pgrid_->leaf_to_level_cells_[this->index()][1];
         return Dune::cpgrid::Entity<0>( *((*(pgrid_ -> level_data_ptr_))[this->level()].get()), entityLevelIdx, true);
     }
-    else {
-        throw std::invalid_argument("The entity provided does not belong to the leaf grid view. ");
-    }
-}
-
-template<int codim>
-Dune::cpgrid::Entity<0> Dune::cpgrid::Entity<codim>::getEquivLevelElem() const
-{
-    // Check if the element belongs to the leaf grid view
-    // This is needed to get the index of the element in the level it was born.
-    // leaf_to_level_cells_ [leaf idx] = {level where the entity was born, equivalent cell idx in that level}
-    if (!(pgrid_ -> leaf_to_level_cells_.empty())) // entity on the LeafGridView
-    {
-        const int& entityLevelIdx = pgrid_->leaf_to_level_cells_[this->index()][1];
-        return Dune::cpgrid::Entity<0>( *((*(pgrid_ -> level_data_ptr_))[this->level()].get()), entityLevelIdx, true);
-    }
     else {
         return *this;
     }
@@ -614,7 +614,6 @@ template<int codim>
 int Dune::cpgrid::Entity<codim>::getLevelCartesianIdx() const
 {
     const auto& level_data = (*(pgrid_ -> level_data_ptr_))[level()].get();
-    // getLevelElem() throws when the entity does not belong to the leaf grid view.
     return level_data -> global_cell_[getLevelElem().index()];
 }
 
