@@ -27,6 +27,8 @@
 #include <opm/grid/utility/platform_dependent/reenable_warnings.h>
 #include <dune/grid/common/mcmgmapper.hh>
 
+#include <numeric>
+
 #if defined(HAVE_ZOLTAN) && defined(HAVE_METIS)
 const int partition_methods[] = {1,2};
 #elif defined (HAVE_ZOLTAN)
@@ -120,11 +122,7 @@ public:
 
     typedef int DataType;
 
-#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 8)
     bool fixedSize()
-#else
-    bool fixedsize()
-#endif
     {
         return true;
     }
@@ -217,11 +215,7 @@ public:
     {}
 
     typedef int DataType;
-#if DUNE_VERSION_NEWER(DUNE_COMMON, 2, 8)
     bool fixedSize()
-#else
-    bool fixedsize()
-#endif
     {
         // We used fixedsize for the message as there is a bug in DUNE
         // up to 2.6.0
@@ -310,7 +304,7 @@ public:
 class CopyCellValues
 {
 public:
-    CopyCellValues(std::vector<int>& cont)
+    explicit CopyCellValues(std::vector<int>& cont)
         : cont_(cont)
     {}
 
@@ -657,9 +651,6 @@ for (auto partition_method : partition_methods) {
         grid.gatherData(gather_gid_set_data);
 
     }
-    decltype(std::get<0>(Dune::CpGrid().loadBalance(nullptr))) test1 = true;
-    decltype(std::get<0>(Dune::CpGrid().loadBalance(Dune::EdgeWeightMethod(), nullptr))) test2 = true;
-    test2 = test1;
 }
 }
 
@@ -684,11 +675,9 @@ BOOST_AUTO_TEST_CASE(distributeParts)
     }
 
     grid.createCartesian(dims, size);
-    std::size_t numCells = 1;
-    for(const auto& dim : dims)
-    {
-        numCells *= dim;
-    }
+    const std::size_t numCells = std::accumulate(dims.begin(), dims.end(), std::size_t{1},
+                                                 [](const auto acc, const auto dim)
+                                                 { return acc*dim; });
 
     auto numCellsPerProc = numCells / grid.comm().size();
     std::vector<int> parts(numCells);
